
#include "windows.h"
#include "Python.h"

HINSTANCE hInstExe;

extern int PyErr_DisplayInDialog( PyObject *exc_type, PyObject *exc_value, PyObject *exc_traceback );
static void CmdLineToArgs(LPTSTR lpCmdLine, int *pargc, char **argv);

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPTSTR lpCmdLine, int nCmdShow)
{	
	PyObject *obhinst;
	BOOL bPythonError = FALSE;
	char *shellModuleName = NULL;
	int argc;
	char *argvbuf[48];
	char **argv=argvbuf;

	hInstExe = hInst;

	SetCursor(LoadCursor(0, IDC_WAIT));

//	Py_pfnPrintError = PyErr_DisplayInDialog;
	Py_InteractiveFlag=1;
	Py_SetProgramName("Python CE");
	Py_Initialize();

	// stash away our hinst.
	obhinst = PyInt_FromLong((long)hInst);
	PySys_SetObject("hinst", obhinst);
	Py_DECREF(obhinst);

/*#ifdef _WIN32_WCE_EMULATION
	argc = __argc-1;
	argv = __argv+1;
#else*/
	CmdLineToArgs(lpCmdLine, &argc, argv);
//#endif

	// check for ID arg, if found, try to find that window.
	if (argc>2 && strcmp(argv[1], "/shell")==0) {
		shellModuleName = argv[2];
		argc -= 2;
		argv += 2;
	}

	// Check for the "/shell" argument
	if (argc>2 && strcmp(argv[1], "/shell")==0) {
		shellModuleName = argv[2];
		argc -= 2;
		argv += 2;
	}
	PySys_SetArgv(argc, argv);

	// A bit of a hack here - if we have a shell module name, then
	// we use PyRun_SimpleFile on it.
	// If we _dont_ have a shell module, we do a simple import
	// of our shell module, and run the main function.
	if (shellModuleName==NULL) {
		// Import our shell module
		PyObject *obMod=PyImport_ImportModule("pcceshell");
		bPythonError = (obMod == NULL);
		if (obMod != NULL) {
			PyObject *obFunc = PyObject_GetAttrString(obMod, "main");
			bPythonError = (obFunc == NULL);
			if (obFunc) {
				PyObject *result = PyObject_CallObject(obFunc, NULL);
				bPythonError = (result==NULL);
				Py_XDECREF(result);
			}
			Py_XDECREF(obFunc);
			Py_DECREF(obMod);
		}
	} else {
		// Run the shell file.
		FILE *f;
		f = Py_fopen(shellModuleName, "r");
		if (f==NULL) {
			char buf[_MAX_PATH+80];
			strcpy(buf, "Can not open the file '");
			strcat(buf, shellModuleName);
			strcat(buf, "'");
			PyErr_SetString(PyExc_ValueError, buf);
			bPythonError = TRUE;
		} 
		else {
			bPythonError = PyRun_SimpleFile(f, shellModuleName) != 0;
			Py_fclose(f);
		}
	}

	if (bPythonError) {
		PyObject *typ, *val, *tb;
		PyErr_Fetch(&typ, &val, &tb);
		PyErr_DisplayInDialog( typ, val, tb );
		Py_XDECREF(typ);
		Py_XDECREF(val);
		Py_XDECREF(tb);
		return FALSE;
	}
	return TRUE;
}





void
CmdLineToArgs(LPTSTR lpCmdLine, int *pargc, char **argv)
{
	BOOL f_skip =TRUE;
	CHAR chQuote='\0';
	CHAR *cmdline=NULL;
	CHAR *p=NULL;

	int argc;
	int charLen = _tcslen(lpCmdLine);
	cmdline = malloc((charLen+1)*sizeof(TCHAR)); // never free'd!
	WideCharToMultiByte(CP_ACP, 0, lpCmdLine, charLen+1, cmdline, charLen+1, NULL, NULL);
	p=cmdline;

	argv[0]="";
	argc=1;

	while (*p)
	{	if (f_skip)
		{
			if (!(*p==' ' || *p=='\t'))
			{
				if (*p=='"' || *p=='\'')
				{	
					chQuote=*p;
					p++;
				}
				argv[argc++]=p;
				f_skip = FALSE;
			}
		}
		else
		{
			if (chQuote) { // We are waiting for the end quote.
				if (*p==chQuote)
				{	
					chQuote='\0';
					f_skip=TRUE;
					*p=0;
				}
			}
			else if ((*p==' ' || *p=='\t'))
			{	
				f_skip=TRUE;
				*p=0;
			}
		}
		p++;
	}
	*pargc=argc;
}
