/* Time module */

#include "Python.h"
#include <tchar.h>

/* --------------------------------------------------------------

   Conversion helpers to make the lack of time_t bearable!

-----------------------------------------------------------------*/

// These UnixTimeTo... functions taken from Knowledge Base Q167296 
DL_IMPORT(BOOL) PyCE_UnixTimeToFileTime(time_t t, LPFILETIME pft)
{
	// Note that LONGLONG is a 64-bit value
	LONGLONG ll;
	ll = ((__int64)t * 10000000) + 116444736000000000;
//	ll = Int32x32To64(t, 10000000) + 116444736000000000;
	pft->dwLowDateTime = (DWORD)ll;
	pft->dwHighDateTime = (DWORD)(ll >> 32);
	return TRUE;
}

DL_IMPORT(BOOL) PyCE_UnixTimeToSystemTime(time_t t, LPSYSTEMTIME pst)
{
	FILETIME ft;

	PyCE_UnixTimeToFileTime(t, &ft);
	return FileTimeToSystemTime(&ft, pst);
}

DL_IMPORT(BOOL) PyCE_FileTimeToUnixTime(FILETIME *pft, time_t *pt)
{
	// Note that LONGLONG is a 64-bit value
	LONGLONG ll = (((__int64)pft->dwHighDateTime)<<32) + pft->dwLowDateTime;
	ll = (ll - 116444736000000000) / 10000000;
	*pt = (time_t)ll;
	return TRUE;
}

DL_IMPORT(BOOL) PyCE_SystemTimeToUnixTime(SYSTEMTIME *pst, time_t *pt)
{
	FILETIME ft;
	if (!SystemTimeToFileTime(pst, &ft))
		return FALSE;
	return PyCE_FileTimeToUnixTime(&ft, pt);
}

DL_IMPORT(void) PyCE_TimeStructToSystemTime(struct tm *ptm, SYSTEMTIME *pst)
{
	pst->wYear = ptm->tm_year;
	if (pst->wYear < 100)
		pst->wYear += 1900;
	pst->wMonth = ptm->tm_mon + 1;
	pst->wDayOfWeek = ptm->tm_wday;
	pst->wDay = ptm->tm_mday;
	pst->wHour = ptm->tm_hour;
	pst->wMinute = ptm->tm_min;
	pst->wSecond = ptm->tm_sec;
	pst->wMilliseconds = 0;
}

DL_IMPORT(double) PyCE_SystemTimeToCTime(SYSTEMTIME* pstTime)
{	SYSTEMTIME stBase;
	FILETIME   ftTime;
	FILETIME   ftBase; 
	__int64    iTime;
	__int64    iBase;

	SystemTimeToFileTime(pstTime, &ftTime);

	stBase.wYear         = 1970;
	stBase.wMonth        = 1;
	stBase.wDayOfWeek    = 1;
	stBase.wDay          = 1;
	stBase.wHour         = 0;
	stBase.wMinute       = 0;
	stBase.wSecond       = 0;
	stBase.wMilliseconds = 0;
	SystemTimeToFileTime(&stBase, &ftBase);

	iTime=ftTime.dwHighDateTime;	
	iTime=iTime << 32;				
	iTime |= ftTime.dwLowDateTime;

	iBase=ftBase.dwHighDateTime;
	iBase=iBase << 32;
	iBase |= ftBase.dwLowDateTime;

	return (double)((iTime - iBase) / 10000000L);
}

// Given a win32 SYSTEMTIME, convert it to a Python time tuple.
static PyObject *
time_convert(SYSTEMTIME *pst)
{	return Py_BuildValue("(iiiiiiiii)", pst->wYear,
										pst->wMonth,
										pst->wDay,
										pst->wHour,
										pst->wMinute,
										pst->wSecond,
										(pst->wDayOfWeek + 6) % 7,
										-1,
										-1);
}

static int
gettmarg(args, p)
	PyObject *args;
	struct tm *p;
{
	if (!PyArg_Parse(args, "(iiiiiiiii)",
			 &p->tm_year,
			 &p->tm_mon,
			 &p->tm_mday,
			 &p->tm_hour,
			 &p->tm_min,
			 &p->tm_sec,
			 &p->tm_wday,
			 &p->tm_yday,
			 &p->tm_isdst))
		return 0;
	if (p->tm_year >= 1900)
		p->tm_year -= 1900;
	p->tm_mon--;
	p->tm_wday = (p->tm_wday + 1) % 7;
	p->tm_yday--;
	return 1;
}

static PyObject *
time_time(PyObject *self, PyObject *args)
{	SYSTEMTIME st;

	if (!PyArg_NoArgs(args))
		return NULL;
	GetSystemTime(&st);
	return PyFloat_FromDouble(PyCE_SystemTimeToCTime(&st));
}


static PyObject *
time_sleep(PyObject *self, PyObject *args)
{
	double d;

	if (!PyArg_Parse(args, "d", &d))
		return NULL;
	Sleep((DWORD)(d * 1000));
	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
time_gmtime(PyObject *self, PyObject *args)
{	SYSTEMTIME st;
	double d;

	if (!PyArg_Parse(args, "d", &d))
		return NULL;

	PyCE_UnixTimeToSystemTime( (time_t)d, &st);
	return time_convert(&st);
}

static PyObject *
time_localtime(PyObject *self, PyObject *args)
{	SYSTEMTIME stLocal;
	FILETIME ftLocal;
	FILETIME ftUTC;
	double d;

	if (!PyArg_Parse(args, "d", &d))
		return NULL;

	PyCE_UnixTimeToFileTime( (time_t)d, &ftUTC);
	FileTimeToLocalFileTime(&ftUTC, &ftLocal);
	FileTimeToSystemTime(&ftLocal, &stLocal);
	return time_convert(&stLocal);
}

const CHAR *dn[]={"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
const CHAR *mn[]={"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", 
				"Aug", "Sep", "Oct", "Nov", "Dec"};

CHAR* wce_ctime(SYSTEMTIME *pst, CHAR* pszString)
{
	int cch = 26; /* hard-coded size of buffer! */
	TCHAR *tempBuf;

	if (/* Day of week already zero based */
		pst->wDayOfWeek<=0 || pst->wDayOfWeek>(sizeof(dn)/sizeof(dn[0])) ||
		/* Month is 1 based */
        pst->wMonth<0 || pst->wMonth>=(sizeof(mn)/sizeof(mn[0]))) {
		*pszString = '\0';
		return pszString;
	}
	// Till we get better Unicode support, use a temp buffer.
	tempBuf = malloc( sizeof(TCHAR) * cch );
	if (tempBuf==NULL) return pszString;

	wsprintf(tempBuf, _T("%hs %hs %2d %2d:%2d:%2d %4d\n"),
		dn[pst->wDayOfWeek], mn[pst->wMonth-1], pst->wDay, 
		pst->wHour, pst->wMinute, pst->wSecond, pst->wYear);

	WideCharToMultiByte(CP_ACP, 0, tempBuf, cch, pszString, cch, NULL, NULL);
	free(tempBuf);

	return (pszString);
}

static PyObject *
time_asctime(self, args)
	PyObject *self;
	PyObject *args;
{
	struct tm buf;
	CHAR s[26]={0};
	SYSTEMTIME st;
	char *p;
	if (!gettmarg(args, &buf))
		return NULL;

	PyCE_TimeStructToSystemTime(&buf, &st);

	p = wce_ctime(&st, s);
	if (p[24] == '\n')
		p[24] = '\0';
	return PyString_FromString(p);
}

static PyObject *
time_ctime(PyObject *self, PyObject *args)
{	CHAR s[26]={0};
	double d;
	CHAR *p;
	SYSTEMTIME st;

	if (!PyArg_Parse(args, "d", &d))
		return NULL;

	PyCE_UnixTimeToSystemTime( (time_t)d, &st);

	p = wce_ctime(&st, s);
	if (p[24] == '\n')
		p[24] = '\0';
	return PyString_FromString(p);
}

static PyObject *
time_clock(self, args)
	PyObject *self;
	PyObject *args;
{
	SYSTEMTIME st;
	GetSystemTime(&st);
	return PyFloat_FromDouble(PyCE_SystemTimeToCTime(&st));
#if 0
	static LARGE_INTEGER ctrStart;
	static LARGE_INTEGER divisor = {0,0};
	LARGE_INTEGER now;
	LONGLONG diff, rem;

	if (!PyArg_NoArgs(args))
		return NULL;

	if (divisor.QuadPart==0) {
		QueryPerformanceCounter(&ctrStart);
		if (!QueryPerformanceFrequency(&divisor) || 
		    (divisor.QuadPart==0)) {
				/* Unlikely to happen - 
				   this works on all intel machines at least! 
				   Revert to time() */
				SYSTEMTIME st;
				GetSystemTime(&st);
				return PyFloat_FromDouble(PyCE_SystemTimeToCTime(&st));
		}
	}
	QueryPerformanceCounter(&now);
	diff = now.QuadPart - ctrStart.QuadPart;
	diff = diff / divisor.QuadPart;
	rem = diff % divisor.QuadPart;
	/* XXX - we assume both divide results fit in 32 bits.  This is
	   true on Intels.  First person who can afford a machine that 
	   doesnt deserves to fix it :-)
	*/
	return PyFloat_FromDouble((double) ((unsigned int)diff + 
		              ((double)((unsigned int)rem)) / (double)(divisor.LowPart)));
#endif
}

static PyObject *
time_mktime(self, args)
	PyObject *self;
	PyObject *args;
{

	struct tm buf;
	time_t t;
	SYSTEMTIME st;
	FILETIME ft;

	if (!gettmarg(args, &buf))
		return NULL;

	PyCE_TimeStructToSystemTime(&buf, &st);	

	if (!SystemTimeToFileTime(&st, &ft)) {
		PyErr_SetString(PyExc_OverflowError,
                                "mktime argument out of range");
		return NULL;
	}

	PyCE_FileTimeToUnixTime(&ft, &t);

	return PyFloat_FromDouble((double)t);
}


static PyMethodDef time_methods[] = {
	{"time",		time_time},
	{"sleep",		time_sleep},
	{"gmtime",		time_gmtime},
	{"localtime",	time_localtime},
	{"asctime",		time_asctime},
	{"ctime",		time_ctime},
	{"clock",       time_clock},
	{"mktime",      time_mktime},
	{NULL, NULL}
};

static void
ins(d, name, v)
	PyObject *d;
	char *name;
	PyObject *v;
{
	if (v == NULL)
		Py_FatalError("Can't initialize time module -- NULL value");
	if (PyDict_SetItemString(d, name, v) != 0)
		Py_FatalError(
		"Can't initialize time module -- PyDict_SetItemString failed");
	Py_DECREF(v);
}

void
inittime()
{
	PyObject *m, *d;
	m = Py_InitModule("time", time_methods);
	d = PyModule_GetDict(m);
#ifdef HAVE_TZNAME
	tzset();
	ins(d, "timezone", PyInt_FromLong((long)timezone));
#ifdef HAVE_ALTZONE
	ins(d, "altzone", PyInt_FromLong((long)altzone));
#else
	ins(d, "altzone", PyInt_FromLong((long)timezone-3600));
#endif
	ins(d, "daylight", PyInt_FromLong((long)daylight));
	ins(d, "tzname", Py_BuildValue("(zz)", tzname[0], tzname[1]));
#else /* !HAVE_TZNAME */
#if HAVE_TM_ZONE
	{
#define YEAR ((time_t)((365 * 24 + 6) * 3600))
		time_t t;
		struct tm *p;
		long winterzone, summerzone;
		char wintername[10], summername[10];
		/* XXX This won't work on the southern hemisphere.
		  XXX Anybody got a better idea? */
		t = (time((time_t *)0) / YEAR) * YEAR;
		p = localtime(&t);
		winterzone = -p->tm_gmtoff;
		strncpy(wintername, p->tm_zone ? p->tm_zone : "   ", 9);
		wintername[9] = '\0';
		t += YEAR/2;
		p = localtime(&t);
		summerzone = -p->tm_gmtoff;
		strncpy(summername, p->tm_zone ? p->tm_zone : "   ", 9);
		summername[9] = '\0';
		ins(d, "timezone", PyInt_FromLong(winterzone));
		ins(d, "altzone", PyInt_FromLong(summerzone));
		ins(d, "daylight",
                    PyInt_FromLong((long)(winterzone != summerzone)));
		ins(d, "tzname",
                    Py_BuildValue("(zz)", wintername, summername));
	}
#else
#ifdef macintosh
	/* The only thing we can obtain is the current timezone
	** (and whether dst is currently _active_, but that is not what
	** we're looking for:-( )
	*/
	initmactimezone();
	ins(d, "timezone", PyInt_FromLong(timezone));
	ins(d, "altzone", PyInt_FromLong(timezone));
	ins(d, "daylight", PyInt_FromLong((long)0));
	ins(d, "tzname", Py_BuildValue("(zz)", "", ""));
#endif /* macintosh */
#endif /* HAVE_TM_ZONE */
#endif /* !HAVE_TZNAME */
#ifdef MS_WINCE
	{  /* Block for variable declarations! */
	TIME_ZONE_INFORMATION tzi;
	GetTimeZoneInformation(&tzi);
	ins(d, "timezone", PyInt_FromLong(tzi.Bias * 60));
	ins(d, "altzone", PyInt_FromLong(tzi.Bias * 60));
	ins(d, "daylight", PyInt_FromLong(tzi.DaylightBias ? 1 : 0));

	// No time zone information for CE.
	ins(d, "tzname", Py_BuildValue("(zz)", "", ""));
	}
#endif /* MS_WINCE */

	if (PyErr_Occurred())
		Py_FatalError("Can't initialize time module");
}


/* Implement floattime() for various platforms */
/* ----- but not for WinCE!!! */
#ifndef MS_WINCE

static double
floattime()
{
	/* There are three ways to get the time:
	  (1) gettimeofday() -- resolution in microseconds
	  (2) ftime() -- resolution in milliseconds
	  (3) time() -- resolution in seconds
	  In all cases the return value is a float in seconds.
	  Since on some systems (e.g. SCO ODT 3.0) gettimeofday() may
	  fail, so we fall back on ftime() or time().
	  Note: clock resolution does not imply clock accuracy! */
#ifdef HAVE_GETTIMEOFDAY
	{
		struct timeval t;
#ifdef GETTIMEOFDAY_NO_TZ
		if (gettimeofday(&t) == 0)
			return (double)t.tv_sec + t.tv_usec*0.000001;
#else /* !GETTIMEOFDAY_NO_TZ */
		if (gettimeofday(&t, (struct timezone *)NULL) == 0)
			return (double)t.tv_sec + t.tv_usec*0.000001;
#endif /* !GETTIMEOFDAY_NO_TZ */
	}
#endif /* !HAVE_GETTIMEOFDAY */
	{
#ifdef HAVE_FTIME
		struct timeb t;
		ftime(&t);
		return (double)t.time + (double)t.millitm * (double)0.001;
#else /* !HAVE_FTIME */
		time_t secs;
		time(&secs);
		return (double)secs;
#endif /* !HAVE_FTIME */
	}
}

#endif
/* Implement floatsleep() for various platforms.
   When interrupted (or when another error occurs), return -1 and
   set an exception; else return 0. */

static int
#ifdef MPW
floatsleep(double secs)
#else
	floatsleep(secs)
	double secs;
#endif /* MPW */
{
/* XXX Should test for MS_WIN32 first! */
#ifdef HAVE_SELECT
	struct timeval t;
	double frac;
	frac = fmod(secs, 1.0);
	secs = floor(secs);
	t.tv_sec = (long)secs;
	t.tv_usec = (long)(frac*1000000.0);
	Py_BEGIN_ALLOW_THREADS
	if (select(0, (fd_set *)0, (fd_set *)0, (fd_set *)0, &t) != 0) {
		Py_BLOCK_THREADS
		PyErr_SetFromErrno(PyExc_IOError);
		return -1;
	}
	Py_END_ALLOW_THREADS
#else /* !HAVE_SELECT */
#ifdef macintosh
#define MacTicks	(* (long *)0x16A)
	long deadline;
	deadline = MacTicks + (long)(secs * 60.0);
	while (MacTicks < deadline) {
		/* XXX Should call some yielding function here */
		if (PyErr_CheckSignals())
			return -1;
	}
#else /* !macintosh */
#if defined(__WATCOMC__) && !defined(__QNX__)
	/* XXX Can't interrupt this sleep */
	Py_BEGIN_ALLOW_THREADS
	delay((int)(secs * 1000 + 0.5));  /* delay() uses milliseconds */
	Py_END_ALLOW_THREADS
#else /* !__WATCOMC__ || __QNX__ */
#ifdef MSDOS
	struct timeb t1, t2;
	double frac;
	extern double fmod Py_PROTO((double, double));
	extern double floor Py_PROTO((double));
	if (secs <= 0.0)
		return;
	frac = fmod(secs, 1.0);
	secs = floor(secs);
	ftime(&t1);
	t2.time = t1.time + (int)secs;
	t2.millitm = t1.millitm + (int)(frac*1000.0);
	while (t2.millitm >= 1000) {
		t2.time++;
		t2.millitm -= 1000;
	}
	for (;;) {
#ifdef QUICKWIN
		Py_BEGIN_ALLOW_THREADS
		_wyield();
		Py_END_ALLOW_THREADS
#endif
		if (PyErr_CheckSignals())
			return -1;
		ftime(&t1);
		if (t1.time > t2.time ||
		    t1.time == t2.time && t1.millitm >= t2.millitm)
			break;
	}
#else /* !MSDOS */
#ifdef MS_WIN32
	/* XXX Can't interrupt this sleep */
	Py_BEGIN_ALLOW_THREADS
	Sleep((int)(secs*1000));
	Py_END_ALLOW_THREADS
#else /* !MS_WIN32 */
#ifdef PYOS_OS2
	/* This Sleep *IS* Interruptable by Exceptions */
	Py_BEGIN_ALLOW_THREADS
	if (DosSleep(secs * 1000) != NO_ERROR) {
		Py_BLOCK_THREADS
		PyErr_SetFromErrno(PyExc_IOError);
		return -1;
	}
	Py_END_ALLOW_THREADS
#else /* !PYOS_OS2 */
	/* XXX Can't interrupt this sleep */
	Py_BEGIN_ALLOW_THREADS
	sleep((int)secs);
	Py_END_ALLOW_THREADS
#endif /* !PYOS_OS2 */
#endif /* !MS_WIN32 */
#endif /* !MSDOS */
#endif /* !__WATCOMC__ || __QNX__ */
#endif /* !macintosh */
#endif /* !HAVE_SELECT */
	return 0;
}
